"""Populate the database and the attachment directory."""

from sys import exit as sys_exit
from os import makedirs, sep
from os.path import exists, join, abspath, basename
from shutil import rmtree
from logging import getLogger
from argparse import REMAINDER
from datetime import datetime

from lxml import etree
from transaction import manager

from ..lib.i18n import _
from ..lib.utils import scandir, copy_content
from ..lib.xml import load_xml, relaxng4validation
from ..lib.config import config_get, config_get_namespace
from ..relaxng import RELAXNG
from ..models import get_tm_dbsession
from ..models.dbsettings import DBSettings
from ..models.dbuser import DBUser
from ..models.populate import xml2db
from . import Script


LOG = getLogger(__name__)


# =============================================================================
def main(args=None):
    """Main function."""
    args = Populate.arguments(Populate.argument_parser(), args)
    if args is not None:
        sys_exit(Populate(args, DBUser, xml2db, RELAXNG).run(args.files))


# =============================================================================
class Populate(Script):
    """Class to populate database.

    :type  args: argparse.Namespace
    :param args:
        Command line arguments.
    :param dbuser_class:
        Class to manage user SQL table.
    :param function _xml2db:
        Function to load XML configuration files. See
        :meth:`~chrysalio.models.populate.xml2db`.
    :param dict relaxng:
        A dictionary describing the main Relax NG with the following keys:
        ``'root'``, ``'version'`` and ``'file'``.
    :param list includes: (optional)
        List of hard coding `includes`.
    :type  dbsession_factory: sqlalchemy.orm.session.sessionmaker
    :param dbsession_factory: (optional)
        Function to create session.
    """
    # pylint: disable = too-many-arguments

    # -------------------------------------------------------------------------
    def __init__(self, args, dbuser_class, _xml2db, relaxng, includes=None,
                 dbsession_factory=None):
        """Constructor method."""
        super(Populate, self).__init__(args, True, includes, dbsession_factory)
        self._dbuser_class = dbuser_class
        self._xml2db = _xml2db
        self._relaxng = relaxng

    # -------------------------------------------------------------------------
    @classmethod
    def argument_parser(cls, description='Populate database.'):
        """Create an argument parser object to parse command line arguments.

        :param str description: (optional)
            Description of the script populating the database.
        :rtype: argparse.ArgumentParser
        """
        parser = super(Populate, cls).argument_parser(description=description)
        parser.add_argument(
            'files', nargs='*', help='optional XML configuration files to use')
        parser.add_argument(
            '--attachments', help='optional attachment directory  to use')
        parser.add_argument(
            '--drop-tables', dest='drop_tables', help='drop existing tables',
            action='store_true')
        parser.add_argument(
            '--remove-locks', dest='remove_locks', action='store_true',
            help='remove locks directory')
        parser.add_argument(
            '--remove-builds', dest='remove_builds', action='store_true',
            help='remove builds directory')
        parser.add_argument(
            '--skip-refresh', dest='skip_refresh', action='store_true',
            help='skip refreshment step')
        parser.add_argument(
            '--recreate-thumbnails', dest='recreate_thumbnails',
            action='store_true', help='recreate thumbnails')
        parser.add_argument(
            '--reindex', dest='reindex', action='store_true',
            help='recreate indexes')
        parser.add_argument('extra', nargs=REMAINDER, help='extra options')
        return parser

    # -------------------------------------------------------------------------
    def run(self, files=None):
        """Check settings and initialize database.

        :param list files: (optional)
            List of files on command-line.
        :rtype: int
        :return:
            Exit code.
        """
        if self.registry is None:
            return 1

        with manager:
            dbsession = get_tm_dbsession(
                self.registry['dbsession_factory'], manager)

            # Load administrator
            LOG.info(self._translate(_('====== Adding administrator')))
            error = self._dbuser_class.load_administrator(
                dbsession, config_get_namespace(
                    self._config, 'Populate', 'admin'))
            if error:
                LOG.error(self._translate(error))
                LOG.critical(self._translate(_('Incorrect administrator.')))
                return 1

            # Load XML configuration files
            self._populate_from_xml(dbsession, files or [])

            # Copy attachments
            self._copy_attachments(self._args.attachments)

            # Update settings
            dbsetting = dbsession.query(DBSettings).filter_by(
                key='populate').first()
            if dbsetting is None:
                dbsetting = DBSettings(key='populate')
                dbsession.add(dbsetting)
            dbsetting.value = datetime.now().isoformat(' ').split('.')[0]

            # Complete operation
            for module_id in self.registry['modules']:
                self.registry['modules'][module_id].activate(
                    self.registry, dbsession)
            for module_id in self.registry['modules']:
                error = self.registry['modules'][module_id].populate(
                    self._args, self.registry, dbsession)
                if error:  # pragma: nocover
                    LOG.error(self._translate(error))
                    return 1

        return 0

    # -------------------------------------------------------------------------
    def _populate_from_xml(self, dbsession, extra_files):
        """Populate database with XML content.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :param list extra_files:
             List of files on command-line.
        """
        # List of files to load
        files = config_get_namespace(self._config, 'Populate', 'data')
        files = [files[k] for k in sorted(files.keys())] + list(extra_files)
        if not files:  # pragma: nocover
            return

        # Load main Relax NG and Relax NG of each module
        relaxngs = relaxng4validation(self._relaxng)
        for module_id in self.registry['modules']:
            if self.registry['modules'][module_id].relaxng is not None:
                relaxngs.update(relaxng4validation(
                    self.registry['modules'][module_id].relaxng))

        # Load files
        LOG.info(self._translate(_('====== Loading configurations')))
        done = set()
        for filename in files:
            filename = abspath(filename)
            if filename in done:
                continue
            LOG.info(filename)
            tree = load_xml(filename, relaxngs)
            # pylint: disable = protected-access
            if not isinstance(tree, etree._ElementTree):
                LOG.error(self._translate(tree))
                continue
            # pylint: enable = protected-access
            errors = self._xml2db(
                dbsession, tree, error_if_exists=False,
                modules=self.registry['modules'])
            for error in errors:  # pragma: nocover
                LOG.error(self._translate(error))
            done.add(filename)

    # -------------------------------------------------------------------------
    def _copy_attachments(self, extra_attachments):
        """Possibly copy attachment files.

        :param list extra_attachments:
             Attachment directory on command-line.
        """
        destination = self.registry.settings.get('attachments')
        if not destination:
            return
        if config_get(self._config, 'Populate', 'drop_tables') == 'true' and \
           exists(destination):
            rmtree(destination)
        if not exists(destination):
            makedirs(destination)

        # List of attachments to copy
        attachments = config_get_namespace(
            self._config, 'Populate', 'attachment')
        attachments = [attachments[k] for k in sorted(attachments.keys())]
        if extra_attachments and exists(extra_attachments):
            if extra_attachments.endswith(sep):
                extra_attachments = extra_attachments[:-len(sep)]
            attachments += [
                join(extra_attachments, k.name)
                for k in scandir(extra_attachments)]
        for attachment in attachments:
            if exists(attachment):
                name = basename(attachment)
                for entry in scandir(attachment):
                    if entry.is_dir():
                        copy_content(
                            entry.path, join(destination, name, entry.name))



"""Various functions to manage attachments."""

from os import makedirs, remove, close
from os.path import join, exists, splitext, basename
from tempfile import mkdtemp, mkstemp


# =============================================================================
def attachment_url(request, category, key, filename):
    """Return the URL of an attachement file or ``None`` if it does not exist
    or attachments are not activated.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str category:
        Category of attachement (e.g. Users).
    :param str key:
        Key to access to the attachments.
    :param str filename:
        Name of file to retrieve.
    :rtype: str
    """
    if not key or not filename:
        return None
    attachments = request.registry.settings.get('attachments')
    if not attachments:
        return None

    attachment = join(attachments, category, key, filename)
    if not exists(attachment):
        return None

    return request.route_path('attachment', path=(category, key, filename))


# =============================================================================
def attachment_update(request, category, key, field, replace=None, prefix=''):
    """Update the content of the attachment directory.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str category:
        Category of attachement (e.g. Users).
    :param str key:
        Key to access to the attachments.
    :type  field: cgi.FieldStorage
    :param field:
        File to retrieve.
    :param str replace: (optional)
        File to delete before adding the new file.
    :param str prefix: (optional)
        Prefix for newly created attachment key.
    :rtype: tuple
    :return:
        A tuple such as ``(attachments_key, file_path)``.
    """
    # Directory for attachments
    attachments = request.registry.settings.get('attachments')
    if not attachments:
        return key, replace
    attachments = join(attachments, category)
    if not exists(attachments):
        try:
            makedirs(attachments)
        except (OSError, IOError):  # pragma: nocover
            request.session.flash(('Permission refused.'), 'alert')
            return key, replace

    # Personal attachment directory
    try:
        if key is None:
            key = basename(mkdtemp(prefix=prefix, dir=attachments))
        attachments = join(attachments, key)
        if not exists(attachments):
            makedirs(attachments)
    except (OSError, IOError):  # pragma: nocover
        request.session.flash(('Permission refused.'), 'alert')
        return key, replace

    # Remove old file
    if replace and exists(join(attachments, replace)):
        try:
            remove(join(attachments, replace))
        except (OSError, IOError):  # pragma: nocover
            request.session.flash(('Permission refused.'), 'alert')
            return key, replace

    # Copy new file
    try:
        hdl, filename = mkstemp(
            suffix=splitext(field.filename)[1], prefix='',
            dir=attachments)
        close(hdl)
        with open(filename, 'wb') as hdl:
            hdl.write(field.file.read())
    except (OSError, IOError):  # pragma: nocover
        request.session.flash(('Permission refused.'), 'alert')
        return key, None

    return key, basename(filename)



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioldap`` classes and functions."""

from unittest import TestCase

from ....tests import ModuleTestCase


# =============================================================================
class UCioLDAPIncludeme(ModuleTestCase):
    """Unit test class for :func:`modules.cioldap.includeme`."""

    # -------------------------------------------------------------------------
    def test_configurator(self):
        """[u:modules.cioldap.includeme] with a configurator"""
        from ....tests import TEST_INI
        from .. import includeme, ModuleCioLDAP

        self.configurator.get_settings().update({
            'ldap.base': 'dc=demo1,dc=freeipa,dc=org',
            'ldap.user.dn':
            'uid=_UID_,cn=users,cn=accounts,dc=demo1,dc=freeipa,dc=org'})

        self.configurator.get_settings()['__file__'] = TEST_INI
        includeme(self.configurator)
        self.assertTrue(self.configurator.registry['modules'])
        self.assertIn(
            'chrysalio.modules.cioldap', self.configurator.registry['modules'])
        self.assertIsInstance(
            self.configurator.registry['modules']['chrysalio.modules.cioldap'],
            ModuleCioLDAP)

    # -------------------------------------------------------------------------
    def test_dictionary(self):
        """[u:modules.cioldap.includeme] with a dictionary"""
        from collections import OrderedDict
        from ....scripts import ScriptRegistry
        from ....tests import TEST_INI
        from .. import includeme, ModuleCioLDAP

        registry = ScriptRegistry({'__file__': TEST_INI})
        registry['modules'] = OrderedDict()

        includeme(registry)
        self.assertEqual(len(registry['modules']), 1)
        self.assertIn('chrysalio.modules.cioldap', registry['modules'])
        self.assertIsInstance(
            registry['modules']['chrysalio.modules.cioldap'], ModuleCioLDAP)


# =============================================================================
class UModuleCioLDAP(TestCase):
    """Unit test class for :class:`modules.cioldap.ModuleCioLDAP`."""

    # -------------------------------------------------------------------------
    def test_activate(self):
        """[u:modules.cioldap.ModuleCioLDAP.activate]"""
        from ....tests import TEST_INI
        from .. import ModuleCioLDAP
        from ..lib.ldap import LDAP
        from ..security import PRINCIPALS_CIOLDAP

        registry = {'principals': []}
        ModuleCioLDAP(TEST_INI).activate(registry, None)
        self.assertIn('principals', registry)
        self.assertIn(PRINCIPALS_CIOLDAP[0], registry['principals'])
        self.assertIn('authorities', registry)
        self.assertIn('ldap', registry['authorities'])
        self.assertIsInstance(registry['authorities']['ldap'], LDAP)

    # -------------------------------------------------------------------------
    def test_deactivate(self):
        """[u:modules.cioldap.ModuleCioLDAP.deactivate]"""
        from ....tests import TEST_INI
        from .. import ModuleCioLDAP
        from ..lib.ldap import LDAP
        from ..security import PRINCIPALS_CIOLDAP

        registry = {
            'principals': [PRINCIPALS_CIOLDAP[0]],
            'authorities': {'ldap': LDAP()}}
        ModuleCioLDAP(TEST_INI).deactivate(registry, None)
        self.assertIn('principals', registry)
        self.assertNotIn(PRINCIPALS_CIOLDAP[0], registry['principals'])
        self.assertIn('authorities', registry)
        self.assertNotIn('ldap', registry['authorities'])



"""Mode management."""

from pyramid.config import Configurator
from pyramid.events import BeforeRender
from pyramid.threadlocal import get_current_request

from .menu import MENU_TEST, MENU_ADMIN, Menu
from .lib.i18n import _
from .lib.modes import Modes

# Mode = (mode_id, (icon, label, permission, route, menu))
MODE_HOME = ('home', (
    '{theme}/images/menu_home.png', _('Home'), None, 'home', None))
MODE_TEST = ('test', (
    '{theme}/images/menu_test1.png', _('Test'), None,
    ('user_view', {'user_id': 3}), MENU_TEST))
MODE_ADMIN = ('admin', (
    '{theme}/images/menu_admin.png', _('Administration'), 'mode-admin',
    'settings_view', MENU_ADMIN))


# =============================================================================
def includeme(configurator):
    """Function to manage modes.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Mode = (mode_id, (label, permission, value_or_function_menu))
    if isinstance(configurator, Configurator):
        configurator.registry['modes'] = [MODE_HOME, MODE_ADMIN]
        configurator.add_subscriber(before_render, BeforeRender)


# =============================================================================
def before_render(event):
    """A subscriber for :class:`pyramid.events.BeforeRender` events to add a
    main menu.

    :type  event: pyramid.events.BeforeRender
    :param event:
        Current event.
    """
    request = event.get('request') or get_current_request()
    event['modes'] = Modes(request, 'modes', request.registry.get('modes', ()))
    menu = Menu(request, 'menu', event['modes'].menu_framework())
    if not menu.is_empty():
        event['menu'] = menu



# pylint: disable = import-outside-toplevel
"""Tests of ``modules.cioldap.models.populate`` functions."""

from ....tests import DBUnitTestCase
from ..models.dbldap import DBLdap


# =============================================================================
class UModelsPopulateXml2Db(DBUnitTestCase):
    """Unit test class for :func:`modules.cioldap.models.populate.xml2db`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modules.cioldap.models.populate.xml2db]"""
        from lxml import etree
        from ..models.populate import xml2db

        root_elt = etree.XML(
            '<cioldap xmlns="http://ns.chrysal.io/cioldap" version="1.0">'
            '  <ldap>'
            '    <base>dc=iinov,dc=com</base>'
            '    <user>'
            '      <dn>uid=_UID_,ou=people,dc=iinov,dc=com</dn>'
            '    </user>'
            '  </ldap>'
            '</cioldap>')
        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member

        self.assertFalse(xml2db(dbsession, root_elt, 'foo', True))
        self.assertEqual(len(xml2db(dbsession, root_elt, None, True)), 0)

        root_elt = etree.XML(
            '<cioldap xmlns="http://ns.chrysal.io/cioldap" version="1.0">'
            '</cioldap>')
        self.assertFalse(xml2db(dbsession, root_elt, 'ldap'))


# =============================================================================
class UModelsPopulateDb2Xml(DBUnitTestCase):
    """Unit test class for :func:`modules.cioldap.models.populate.db2xml`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:modules.cioldap.models.populate.db2xml]"""
        from lxml import etree
        from ..models.populate import db2xml

        # pylint: disable = no-member
        dbsession = self.request.dbsession
        # pylint: enable = no-member
        dbsession.add(DBLdap(
            base='dc=iinov,dc=fr',
            user_dn='uid=_UID_,ou=people,dc=iinov,dc=com',
            user_filter='(&amp;(objectclass=inetOrgPerson)(uid=_UID_))',
            field_first_name='givenName',
            field_last_name='sn',
            field_email='mail'))
        namespace = 'http://ns.chrysal.io/cioldap'
        root_elt = etree.Element(
            '{{{0}}}cioldap'.format(namespace),
            nsmap={None: namespace}, version='1.0')

        db2xml(dbsession, root_elt)
        self.assertEqual(len(root_elt), 1)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.breadcrumbs`` class methods."""

from unittest import TestCase


# =============================================================================
class ULibBreadcrumbsBreadcrumbs(TestCase):
    """Unit test class for :class:`lib.breadcrumbs.Breadcrumbs`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from pyramid.testing import setUp, DummyRequest

        self.configurator = setUp()
        self.request = DummyRequest(matched_route=None)

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from pyramid.testing import tearDown
        tearDown()

    # -------------------------------------------------------------------------
    def make_one(self):
        """Create an instance of breadcrumbs.

        rtype: Breadcrumbs
        """
        from ..lib.breadcrumbs import Breadcrumbs

        self.configurator.add_route('home', '/')
        self.configurator.add_route('user_index', '/user/index')
        self.configurator.add_route('user_edit', '/user/edit')
        self.configurator.add_route('profile_index', '/profile/index')
        return Breadcrumbs(self.request)

    # -------------------------------------------------------------------------
    def test_call_home(self):
        """[u:lib.breadcrumbs.Breadcrumbs.__call__] home"""
        from collections import namedtuple

        breadcrumbs = self.make_one()
        breadcrumbs('Home', 1)
        self.assertIn('breadcrumbs', self.request.session)
        crumbs = self.request.session['breadcrumbs']
        self.assertEqual(len(crumbs), 1)
        self.assertEqual(crumbs[0][0], 'Home')
        self.assertIsNone(crumbs[0][1])
        self.assertEqual(crumbs[0][2], {})
        self.assertEqual(crumbs[0][3], 20)

        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        breadcrumbs('Home', 0)
        crumbs = self.request.session['breadcrumbs']
        self.assertEqual(crumbs[0][1], 'home')

    # -------------------------------------------------------------------------
    def test_call_replace(self):
        """[u:lib.breadcrumbs.Breadcrumbs.__call__] with replacement"""
        from collections import namedtuple

        breadcrumbs = self.make_one()
        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        breadcrumbs('Home', 1)
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')
        breadcrumbs('User List')
        self.assertEqual(len(self.request.session['breadcrumbs']), 2)
        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_index')
        breadcrumbs('Profile List', replace=breadcrumbs.current_path())
        self.assertEqual(len(self.request.session['breadcrumbs']), 2)
        breadcrumbs('Profile List')
        self.assertEqual(len(self.request.session['breadcrumbs']), 2)

    # -------------------------------------------------------------------------
    def test_call_anchor(self):
        """[u:lib.breadcrumbs.Breadcrumbs.__call__] with anchor"""
        breadcrumbs = self.make_one()
        breadcrumbs('Home', forced_route=('home', {}))
        breadcrumbs(
            'User Edition', anchor='current',
            forced_route=('user_index', {'user_id': 0}))
        self.assertEqual(len(self.request.session['breadcrumbs']), 2)
        last_crumb = self.request.session['breadcrumbs'][-1]
        self.assertEqual(len(last_crumb[2]), 2)
        self.assertIn('user_id', last_crumb[2])
        self.assertIn('_anchor', last_crumb[2])
        self.assertEqual(last_crumb[2]['user_id'], 0)
        self.assertEqual(last_crumb[2]['_anchor'], 'current')

    # -------------------------------------------------------------------------
    def test_pop(self):
        """[u:lib.breadcrumbs.Breadcrumbs.pop]"""
        breadcrumbs = self.make_one()
        breadcrumbs('Home', forced_route=('home', {}))
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(len(self.request.session['breadcrumbs']), 2)
        breadcrumbs.pop()
        self.assertEqual(len(self.request.session['breadcrumbs']), 1)
        self.assertEqual(self.request.session['breadcrumbs'][0][1], 'home')

    # -------------------------------------------------------------------------
    def test_trail(self):
        """[u:lib.breadcrumbs.Breadcrumbs.trail]"""
        from webhelpers2.html import literal

        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.trail(), literal('&nbsp;'))
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.trail(), literal('&nbsp;'))
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(
            breadcrumbs.trail(), literal('<a href="/">Home</a> ‣ User List'))
        breadcrumbs.pop()
        breadcrumbs('User List')
        breadcrumbs('Profile List')
        self.assertEqual(
            breadcrumbs.trail(),
            literal('<a href="/">Home</a> ‣ User List ‣ Profile List'))

    # -------------------------------------------------------------------------
    def test_current_title(self):
        """[u:lib.breadcrumbs.Breadcrumbs.current_title]"""
        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.current_title(), 'Home')
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.current_title(), 'Home')
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(breadcrumbs.current_title(), 'User List')

    # -------------------------------------------------------------------------
    def test_current_route_name(self):
        """[u:lib.breadcrumbs.Breadcrumbs.current_route_name]"""
        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.current_route_name(), 'home')
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.current_route_name(), 'home')
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(breadcrumbs.current_route_name(), 'user_index')

    # -------------------------------------------------------------------------
    def test_current_path(self):
        """[u:lib.breadcrumbs.Breadcrumbs.current_path]"""
        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.current_path(), '/')
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.current_path(), '/')
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(breadcrumbs.current_path(), '/user/index')

    # -------------------------------------------------------------------------
    def test_crumb_trail(self):
        """[u:lib.breadcrumbs.Breadcrumbs.crumb_trail]"""
        from collections import namedtuple
        from ..lib.breadcrumbs import DEFAULT_ROOT_CHUNKS

        breadcrumbs = self.make_one()
        self.assertEqual(len(breadcrumbs.crumb_trail()), 0)

        self.request.matched_route = namedtuple('Route', 'name')(name='home')
        crumb_trail = breadcrumbs.crumb_trail()
        self.assertEqual(len(crumb_trail), 1)
        self.assertEqual(crumb_trail[0], ([''], DEFAULT_ROOT_CHUNKS))

        breadcrumbs('Home', forced_route=('home', {}))
        crumb_trail = breadcrumbs.crumb_trail()
        self.assertEqual(len(crumb_trail), 1)
        self.assertEqual(crumb_trail[0], ([''], DEFAULT_ROOT_CHUNKS))

        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')
        breadcrumbs('User Management', forced_route=('user_index', {}))
        self.assertEqual(len(breadcrumbs.crumb_trail()), 2)

        self.request.matched_route = namedtuple('Route', 'name')(
            name='profile_index')
        self.assertEqual(len(breadcrumbs.crumb_trail()), 3)

    # -------------------------------------------------------------------------
    def test_back_title(self):
        """[u:lib.breadcrumbs.Breadcrumbs.back_title]"""
        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.back_title(), 'Home')
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.back_title(), 'Home')
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(breadcrumbs.back_title(), 'Home')
        breadcrumbs('Profile List', forced_route=('user_profile', {}))
        self.assertEqual(breadcrumbs.back_title(), 'User List')

    # -------------------------------------------------------------------------
    def test_back_path(self):
        """[u:lib.breadcrumbs.Breadcrumbs.back_path]"""
        breadcrumbs = self.make_one()
        self.assertEqual(breadcrumbs.back_path(), '/')
        breadcrumbs('Home', forced_route=('home', {}))
        self.assertEqual(breadcrumbs.back_path(), '/')
        breadcrumbs('User List', forced_route=('user_index', {}))
        self.assertEqual(breadcrumbs.back_path(), '/')
        breadcrumbs('Profile List', forced_route=('profile_index', {}))
        self.assertEqual(breadcrumbs.back_path(), '/user/index')



"""Panel class."""

from webhelpers2.html import literal
from chameleon import PageTemplateFile

from pyramid.asset import abspath_from_asset_spec

from chrysalio.lib.utils import tostr
from chrysalio.lib.form import get_action, Form
from ..includes.themes import theme_static_prefix
from .i18n import _


PANEL_ITEM_PREFIX = 'panel:'


# =============================================================================
class Panel(object):
    """Class to manage side panel.

    :param list area: (optional)
        List of route names which determine the area where the panel is
        visible.
    """

    uid = None
    label = None
    icon = '/images/panel.png'
    template = 'chrysalio:Templates/panel.pt'
    css = ()
    javascripts = ()
    constants = {}
    need_form = False

    # -------------------------------------------------------------------------
    def __init__(self, area=None):
        """Constructor method."""
        self.area = area or ()

    # -------------------------------------------------------------------------
    @classmethod
    def register(cls, registry, panel_class, area=None, add2systray=True):
        """Method to register the panel and possibly add it to the panel
        systray.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :param panel_class:
            Panel class.
        :param list area: (optional)
            List of route names which determine the area where the panel is
            visible.
        :param bool add2systray: (default=True)
            If ``True`` add the panel to systray.
        :rtype: :class:`.lib.panel.Panel` or ``None``
        """
        if panel_class.uid is None or panel_class.label is None:
            return None
        if 'panels' not in registry:
            registry['panels'] = {}
        if panel_class.uid in registry['panels']:
            return registry['panels'][panel_class.uid]
        registry['panels'][panel_class.uid] = panel_class(area)

        if add2systray:
            if 'systray' not in registry:
                registry['systray'] = []
            registry['systray'].append(registry['panels'][panel_class.uid])

        return registry['panels'][panel_class.uid]

    # -------------------------------------------------------------------------
    @classmethod
    def has_open_panel(cls, request):
        """Return ``'cioHasOpenPanel'`` if almost one panel is open.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: class:`str` or ``None``
        """
        if 'panels' not in request.registry or 'panels' not in request.session:
            return None
        for panel in request.registry['panels'].values():
            if panel.uid in request.session['panels'] and \
               request.session['panels'][panel.uid]['is_open']:
                return 'cioHasOpenPanel'
        return None

    # -------------------------------------------------------------------------
    def is_open(self, request):
        """``True`` if the panel is open.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        self._prepare_session(request)
        if self.area and (
                request.matched_route is None or
                request.matched_route.name not in self.area):
            self.close(request)
        return request.session['panels'][self.uid]['is_open']

    # -------------------------------------------------------------------------
    def was_open(self, request):
        """``True`` if the panel was previously open.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        self._prepare_session(request)
        return request.session['panels'][self.uid]['was_open']

    # -------------------------------------------------------------------------
    def open(self, request):
        """Open the panel and memorize the state.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        self._prepare_session(request)
        request.session['panels'][self.uid]['was_open'] = self.is_open(request)
        request.session['panels'][self.uid]['is_open'] = True

        if 'panels' in request.registry:
            for other in request.registry['panels'].values():
                if other.uid != self.uid:
                    other.close(request)

    # -------------------------------------------------------------------------
    def close(self, request):
        """Close the panel and memorize the state.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        self._prepare_session(request)
        request.session['panels'][self.uid]['was_open'] = False
        request.session['panels'][self.uid]['is_open'] = False

    # -------------------------------------------------------------------------
    def clear_values(self, request):
        """Clear all values of this panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        if 'panels' in request.session \
           and self.uid in request.session['panels'] \
           and 'values' in request.session['panels'][self.uid]:
            del request.session['panels'][self.uid]['values']

    # -------------------------------------------------------------------------
    def set_values(self, request, values):
        """Set values of this panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param dict values:
            Values to set.
        """
        self._prepare_session(request)
        request.session['panels'][self.uid]['values'] = values

    # -------------------------------------------------------------------------
    def values(self, request):
        """Return values of this panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: dict
        """
        self._prepare_session(request)
        return request.session['panels'][self.uid].get('values', {})

    # -------------------------------------------------------------------------
    def set_value(self, request, value_id, value):
        """Set a value of this panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str value_id:
            Value ID.
        :param value:
            Value to set.
        """
        self._prepare_session(request)
        if 'values' not in request.session['panels'][self.uid]:
            request.session['panels'][self.uid]['values'] = {}
        request.session['panels'][self.uid]['values'][value_id] = value

    # -------------------------------------------------------------------------
    def value(self, request, value_id):
        """Return the value ``value_id`` of this panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str value_id:
            ID of the requested value.
        :rtype: str
        """
        return self.values(request).get(value_id)

    # -------------------------------------------------------------------------
    def render(self, request, ts_factory=_):
        """Return the content of the panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param ts_factory: (default=_)
            Translation String Factory fucntion.
        :rtype: webhelpers2.html.literal
        """
        domain = self.template.partition(':')[0] \
            if ':' in self.template else 'chrysalio'

        def _translate(
                msgid, domain=domain, mapping=None, target_language=None):
            """Translation for Chameleon."""
            # pylint: disable = unused-argument
            return request.localizer.translate(
                msgid, domain=domain, mapping=mapping)

        params = {
            'request': request,
            '_':  ts_factory,
            'route': lambda name, *elts, **kwargs: tostr(
                request.route_path(name, *elts, **kwargs)),
            'theme': theme_static_prefix(request),
            'get_action': lambda request: get_action(request, True),
            'PANEL_ITEM_PREFIX': PANEL_ITEM_PREFIX,
            'panel': self}
        params.update(self.constants)
        params.update(self.values(request))
        if self.need_form:
            params['form'] = Form(
                request, defaults=params.get('form_defaults'),
                force_defaults='form_defaults' in params)
            params['form'].forget(PANEL_ITEM_PREFIX)

        return literal(PageTemplateFile(
            abspath_from_asset_spec(self.template),
            translate=_translate).render(**params))

    # -------------------------------------------------------------------------
    def route(self, request):
        """Return the route to toggle the state of the panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: str
        """
        if 'panel' in request.GET:
            del request.GET['panel']
        query_string = dict(request.GET)
        query_string.update({'panel': self.uid})
        if request.matched_route is None:
            self.close(request)
            return ''
        return request.current_route_path(_query=query_string)

    # -------------------------------------------------------------------------
    @classmethod
    def open_panel_css(cls, request):
        """Return a list of URL of CSS used by the open panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: tuple
        """
        for panel in request.registry.get('panels', {}).values():
            if panel.is_open(request):
                return panel.css

        return ()

    # -------------------------------------------------------------------------
    @classmethod
    def open_panel_js(cls, request):
        """Return a list of URL of Javascript used by the open panel.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: tuple
        """
        for panel in request.registry.get('panels', {}).values():
            if panel.is_open(request):
                return panel.javascripts

        return ()

    # -------------------------------------------------------------------------
    @classmethod
    def manage_panels(cls, request):
        """Possibly, toggle the current panel state.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        if 'panels' not in request.registry:
            return

        # Was open
        for panel in request.registry['panels'].values():
            if panel.is_open(request):
                request.session['panels'][panel.uid]['was_open'] = True

        # Close or open a panel
        panel_id = request.GET.get('panel')
        if panel_id and panel_id in request.registry['panels']:
            panel = request.registry['panels'][panel_id]
            if panel.is_open(request):
                panel.close(request)
            else:
                panel.open(request)

    # -------------------------------------------------------------------------
    def _prepare_session(self, request):
        """Prepare session to memorize panel state.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        """
        if 'panels' not in request.session:
            request.session['panels'] = {}
        if self.uid not in request.session['panels']:
            request.session['panels'][self.uid] = {
                'was_open': False, 'is_open': False}



"""Form validation and rendering library."""

from re import sub as re_sub

import colander
from webhelpers2.html import tags, HTML, literal

from pyramid.csrf import get_csrf_token, new_csrf_token

from .i18n import _


# =============================================================================
def get_action(request, silent=False):
    """Return a tuple such as ``(action, items)`` where ``action`` is a
    string such as ``<act><?|!><item_id>`` and ``items`` is a list of
    selected items in a list form.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param bool silent: (default=False)
        If ``True``, no alert emitted.
    :rtype: tuple
    :return: (tuple)
        A tuple such as ``(action, items)``.

    Each submit button returns a string such as ``<act><?|!><item_id>.x`` where
    ``<item_id>`` is the item identifier or ``#`` for all selected items,
    ``<?|!>`` means respectively *confirm* or *proceed* and ``<act>`` is the
    action to do.

    Checkbox inputs return string such as ``#<item_id>``.

    For instance, ``del!#`` and ``['#user1', '#user2']`` means "delete
    ``user1`` and ``user2``". ``del!user1`` means "delete ``user1`` and only
    this one".
    """
    action = ''
    items = []
    for param in request.POST:
        if not isinstance(param, str):
            param = param.decode('utf8')  # pragma: nocover
        if param[0] == '#':
            items.append(param[1:])
        elif param[-2:] == '.x':
            if param[-3:-2] == '#':
                action = param[0:-2]
            else:
                return param[0:-2], (param[4:-2],)
    if '#' in action and not items:
        if not silent:
            request.session.flash(_('Select items!'), 'alert')
        action = ''
    return action, items


# =============================================================================
class SameAs(object):
    # pylint: disable = too-few-public-methods
    """This class implements a ``colander`` validator to check if to fields are
    identical.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str reference:
        Name of the field to compare with.
    :type  message: :class:`str` or :class:`~pyramid.i18n.TranslationString`
    :param message: (optional)
        Error message.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request, reference, message=None):
        """Constructor method."""
        self._request = request
        self._reference = reference
        self._message = message or _('The two fields are not identical.')

    # -------------------------------------------------------------------------
    def __call__(self, node, value):
        """This method raises a :class:`colander.Invalid` instance as an
        exception value is not same as ``self.reference``.

        :type node: colander.SchemaNode
        :type value: cstruct
        """
        if self._request.POST.get(self._reference) != value:
            raise colander.Invalid(node, self._message)


# =============================================================================
def button(url, label='', src=None, title=None, class_='cioButton'):
    """Output a link on a label and an image with a button aspect.

    :param str url:
        Target URL.
    :param str label: (optional)
        Label for roll over and ``alt``.
    :param str src: (optional)
        Image path.
    :param str title: (optional)
        Label for roll over.
    :param str class_: (default='cioButton')
        The class attribute.
    :rtype: str
        HTML tag.
    """
    if class_ == 'cioButton' and not label and src:
        class_ = None
    return literal('<a href="{0}"{1}{2}>{3}{4}</a> '.format(
        literal.escape(url), ' title="{0}"'.format(title) if title else '',
        ' class="{0}"'.format(class_) if class_ else '',
        src and '<img src="%s" alt="%s"/>' % (src, label or title) or '',
        label))


# =============================================================================
def grid_item(name, label, content, required=False, hint=None, error=None,
              title=None, clear=False, class_=None):
    """Display an item with label, hint and error message.

    :param str name:
        Input ID.
    :param str label:
        Label.
    :param str content:
        HTML content.
    :param bool required: (default=False)
        Indicate if this field is required.
    :param str hint: (optional)
        Help message.
    :param str error: (optional)
        Error message.
    :param str title: (optional)
        Title for the hover effect.
    :param bool clear: (default=False)
        If ``True``, add a ``<div class="cioClear"/>`` at the end.
    :param str class_: (optional)
        The class attribute.
    :rtype: str

    This ouputs a structure such as:

    .. code-block:: html

        <div class="[class_]">
          <label for="[name]"><strong>[label]<span>*</span></strong></label>
          <tag title="[title]">
            [content]
            <em> [hint]</em>
            <strong> [form.error(name)]</strong>
          </tag>
          <div class="cioClear"></div>
        </div>
    """
    # pylint: disable = too-many-arguments
    if not content:
        return ''
    if error:
        class_ = '{0} cioError'.format(class_) if class_ else 'cioError'
    return literal(
        '<div{class_}>'
        '<label{name}><strong>{label}{required}</strong></label>'
        '<div{title}>{content}{hint}{error}</div>{clear}</div>'.format(
            class_=' class="{0}"'.format(class_) if class_ else '',
            name=' for="{0}"'.format(name.replace(':', '')) if name else '',
            label=label or '',
            required=HTML.span('*') if required else '',
            title=' title="{0}"'.format(title) if title else '',
            content=content,
            hint=HTML.em(' {0}'.format(hint)) if hint else '',
            error=HTML.strong(' {0}'.format(error)) if error else '',
            clear=clear and '<div class="cioClear"></div>' or ''))


# =============================================================================
class Form(object):
    """Form validation class."""
    # pylint: disable = too-many-public-methods

    # -------------------------------------------------------------------------
    def __init__(self, request, schema=None, defaults=None, obj=None,
                 force_defaults=False):
        """Constructor method."""
        # pylint: disable = too-many-arguments
        self.values = defaults \
            if defaults and (not request.POST or force_defaults) else {}
        self._request = request
        self._schema = schema
        self._errors = {}
        self._special = [[], []]
        self._validated = False
        if obj is not None and schema is not None and not request.POST:
            for field in [k.name for k in schema]:
                if hasattr(obj, field):
                    self.values[field] = getattr(obj, field)

    # -------------------------------------------------------------------------
    def validate(self, obj=None):
        """Check if the form is validated.

        :param object obj: (optional)
            Object to fill.
        :rtype: bool
        :return:
             ``True`` if validated.
        """
        # Something to do?
        if not self._request.POST:
            return False
        if self._validated:
            return not self._errors

        # Schema validation
        params = dict(self._request.POST.items())
        if self._schema:
            try:
                self.values = self._schema.deserialize(params)
            except colander.Invalid as err:
                self._errors = {}
                for child in err.children:
                    self._errors[child.node.name] = child.messages()
        else:
            self.values.update(params)

        # Fill object
        if obj is not None and not self._errors:
            for field in self.values:
                if hasattr(obj, field):
                    setattr(obj, field, self.values[field])

        self._validated = True
        return len(self._errors) == 0

    # -------------------------------------------------------------------------
    def has_error(self, name=None):
        """Return ``True`` if field ``name`` has an error.

        :param str name: (optional)
            Input ID.
        :rtype: bool
        """
        return bool(name is None and self._errors) or name in self._errors

    # -------------------------------------------------------------------------
    def set_error(self, name, message):
        """Set an error message for field ``name``.

        :param str name:
            Input ID.
        :param str message:
            Error message.
        """
        if name in self._errors:
            self._errors[name].append(message)
        else:
            self._errors[name] = [message]

    # -------------------------------------------------------------------------
    def error(self, name):
        """Return error message for field ``name``.

        :param str name:
            Input ID.
        :rtype: str
        :return:
            Translated error message.
        """
        if name not in self._errors:
            return ''
        return ' ; '.join([self._request.localizer.translate(error)
                           for error in self._errors[name]])

    # -------------------------------------------------------------------------
    def static(self, name):
        """The field ``name`` will not be updated by the form.

        :param str name:
            Name of field to set static.
        """
        if name not in self._special[0]:
            self._special[0].append(name)

    # -------------------------------------------------------------------------
    def forget(self, prefix):
        """Fields beginning by ``prefix`` are forgotten when the page is
        refreshed.

        :param str prefix:
            Prefix to select fields.
        """
        if prefix not in self._special[1]:
            self._special[1].append(prefix)

    # -------------------------------------------------------------------------
    @classmethod
    def make_safe_id(cls, name):
        """Make a string safe for including in an id attribute

        :param str name:
            String to transform.
        :rtype: str
        """
        return re_sub(r'(?!-)\W', '', re_sub(
            r'\s', '_', name.replace('?', '0').replace('!', '1'), )).lower()

    # -------------------------------------------------------------------------
    def begin(self, url=None, multipart=False, **attrs):
        """Ouput the ``<form>`` tag.

        :param str url: (optional)
            URL to submit form, by default, the current URL.
        :param bool multipart: (default=False)
            If set to ``True``, the enctype is set to ``multipart/form-data``.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        token = get_csrf_token(self._request) or new_csrf_token(self._request)
        html = tags.form(
            url or self._request.path_qs, 'post', multipart, **attrs)
        html += HTML.div(self.hidden('csrf_token', token), class_="cioHidden")
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def end(cls):
        """Ouput the ``</form>`` tag."""
        return tags.end_form()

    # -------------------------------------------------------------------------
    @classmethod
    def submit(cls, name, label=None, class_='cioButton', **attrs):
        """Output a submit button with the label as the caption.

        :param str name:
            Input ID.
        :param str label: (optional)
            Button caption.
        :param str class_: (default='cioButton')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return tags.submit(
            name, label, cls.make_safe_id(name), class_=class_, **attrs)

    # -------------------------------------------------------------------------
    @classmethod
    def submit_image(cls, name, label, src, **attrs):
        """Output an image submit button.

        :param str name:
            Input ID.
        :param str label:
            Label for roll over and ``alt``.
        :param str src:
            Image path.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return HTML.tag(
            'input', type='image', name=name, id=cls.make_safe_id(name),
            src=src, title=label or name, alt=label or name, **attrs)

    # -------------------------------------------------------------------------
    @classmethod
    def submit_cancel(cls, label, src, **attrs):
        """Output a cancel submit button.

        :param str label:
            Label for roll over and ``alt``.
        :param str src:
            Image path.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return HTML.tag(
            'input', type='image', name='ccl!', id='ccl', src=src, title=label,
            alt=label, **attrs)

    # -------------------------------------------------------------------------
    @classmethod
    def button(cls, url, label='', src=None, title=None, class_='cioButton'):
        """Output a link on a label and an image with a button aspect.

        See :func:`button`.
        """
        return button(url, label, src, title, class_)

    # -------------------------------------------------------------------------
    @classmethod
    def default_button(cls, name):
        """Create an invisible button to catch the Enter signal for the input
        submit we want to be the default one.

        :param str name:
            ID of the input submit to activate.
        """
        return literal(
            '<button type="submit" name="{0}" class="cioDefaultButton">'
            '</button>'.format(name))

    # -------------------------------------------------------------------------
    @classmethod
    def grid_item(cls, label, content, required=False, hint=None, error=None,
                  title=None, clear=False, class_='cioFormItem'):
        """Output an item with label, hint and error message.

        See :func:`grid_item`.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            None, label, content, required, hint, error, title, clear,
            class_)

    # -------------------------------------------------------------------------
    def hidden(self, name, value=None, **attrs):
        """Output a hidden field.

        :param str name:
            Input ID.
        :param str value: (optional)
            Hidden value.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return tags.hidden(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def text(self, name, value=None, **attrs):
        """Output a standard text field.

        :param str name:
            Input ID.
        :param str value: (optional)
            Default value.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
             HTML tag.
        """
        return tags.text(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def password(self, name, value=None, **attrs):
        """Output a password field.

        This method takes the same options as text().
        """
        return tags.password(name, self._value(name, value), **attrs)

    # -------------------------------------------------------------------------
    def checkbox(self, name, value='1', checked=False, autosubmit=False,
                 **attrs):
        """Output a check box.

        :param str name:
            Input ID.
        :param str value: (default='1')
            The value to return to the application if the box is checked.
        :param bool checked: (default=False)
            ``True`` if the box should be initially checked.
        :param bool autosubmit: (default=False)
            If ``True``, it adds ``class="cioAutoSubmit"`` attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        if autosubmit:
            attrs['class_'] = '{0} cioAutoSubmit'.format(attrs['class_']) \
                if 'class_' in attrs else 'cioAutoSubmit'
        return tags.checkbox(
            name, value, checked or self._value(name), **attrs)

    # -------------------------------------------------------------------------
    def custom_checkbox(
            self, name, value='1', checked=False, autosubmit=False,
            class_=None, **attrs):
        """Output a check box followed by an empty label to customize the
        aspect of the box.

        :param str name:
            Input ID.
        :param str value: (default='1')
            The value to return to the application if the box is checked.
        :param bool checked: (default=False)
            ``True`` if the box should be initially checked.
        :param bool autosubmit: (default=False)
            If ``True``, it adds ``class="cioAutoSubmit"`` attribute.
        :param str class_: (default='cioCustomCheckbox')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        if class_ is None:
            class_ = 'cioCustomCheckbox'
        attrs['class_'] = class_
        if autosubmit:
            attrs['class_'] = '{0} cioAutoSubmit'.format(attrs['class_'])
        return literal('{0}{1}'.format(
            tags.checkbox(name, value, checked or self._value(name), **attrs),
            '<label for="{0}" class="{1}"> </label>'.format(
                self.make_safe_id(name), class_)))

    # -------------------------------------------------------------------------
    def radio(self, name, value, checked=False, autosubmit=False, **attrs):
        """Output a radio button.

        :param str name:
            Input ID.
        :param str value:
            The value to return to the application if the radio is checked.
        :param bool checked: (default=False)
            ``True`` if the box should be initially checked.
        :param bool autosubmit: (default=False)
            If ``True``, it adds ``class="cioAutoSubmit"`` attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        if autosubmit:
            attrs['class_'] = '{0} cioAutoSubmit'.format(attrs['class_']) \
                if 'class_' in attrs else 'cioAutoSubmit'
        return tags.radio(
            name, value, checked or value == self._value(name), **attrs)

    # -------------------------------------------------------------------------
    def select(self, name, selected_values, options, autosubmit=False,
               **attrs):
        """Output a dropdown selection box.

        :param str name:
            Input ID.
        :type  selected_value: :class:`str` or :class:`list`
        :param selected_value:
            A string or list of strings or integers giving the value(s) that
            should be preselected.
        :type options:
            (list of :class:`str`, :class:`int` or ``(value, label)`` pairs)
        :param options:
            The label will be shown on the form; the option will be returned to
            the application if that option is chosen. If you pass a ``string``
            or ``int`` instead of a ``2-tuple``, it will be used for both the
            value and the label.
        :param bool autosubmit: (default=False)
            If ``True``, it adds ``class="cioAutoSubmit"`` attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        if not options:
            return ''
        opts = []
        translate = self._request.localizer.translate
        for opt in options:
            if isinstance(opt, tuple):
                opts.append(
                    tags.Option(translate(opt[1]), '{0}'.format(opt[0])))
            else:
                opts.append(tags.Option('{0}'.format(opt)))
        if autosubmit:
            attrs['class_'] = '{0} cioAutoSubmit'.format(attrs['class_']) \
                if 'class_' in attrs else 'cioAutoSubmit'
        return tags.select(
            name, '{0}'.format(self._value(name, selected_values)),
            tags.Options(opts), **attrs)

    # -------------------------------------------------------------------------
    def upload(self, name, value=None, **attrs):
        """Output a file upload field.

        :param str name:
            Input ID.
        :param str value: (optional)
            Default value.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return tags.file(name, self._value(name, value) or None, **attrs)

    # -------------------------------------------------------------------------
    def textarea(self, name, content='', **attrs):
        """Output a text input area.

        :param str name:
            Input ID.
        :param str content: (optional)
            Default value.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            HTML tag.
        """
        return tags.textarea(name, self._value(name, content), **attrs)

    # -------------------------------------------------------------------------
    def grid_text(self, name, label, required=False, hint=None, title=None,
                  clear=False, class_='cioFormItem', **attrs):
        """Output a standard text field in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.text(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def grid_password(self, name, label, required=False, hint=None, title=None,
                      clear=False, class_='cioFormItem', **attrs):
        """Output a password field in a CSS grid layout.

        This method takes the same options as grid_text().
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.password(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def grid_checkbox(self, name, label, required=False, hint=None, title=None,
                      clear=False, class_='cioFormItem', **attrs):
        """Output a check box in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.checkbox(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def grid_custom_checkbox(
            self, name, label, required=False, hint=None, title=None,
            clear=False, class_='cioFormItem', **attrs):
        """Output a custom check box in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.custom_checkbox(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def grid_select(self, name, label, options, autosubmit=False,
                    required=False, hint=None, title=None, clear=False,
                    class_='cioFormItem', **attrs):
        """Output a dropdown selection box in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :type options:
            (list of :class:`str`, :class:`int` or ``(value, label)`` pairs)
        :param options:
            Values in the dropdown list.
        :param bool autosubmit: (default=False)
            If ``True``, it adds ``onchange="submit()"`` attribute.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        if not options:
            return ''
        return grid_item(
            name, label, self.select(name, None, options, autosubmit, **attrs),
            required, hint, self.error(name), title=title, clear=clear,
            class_=class_)

    # -------------------------------------------------------------------------
    def grid_upload(self, name, label, required=False, hint=None, title=None,
                    clear=False, class_='cioFormItem', **attrs):
        """Output a file upload field in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.upload(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def grid_textarea(self, name, label, required=False, hint=None, title=None,
                      clear=False, class_='cioFormItem', **attrs):
        """Output a text input area in a CSS grid layout.

        :param str name:
            Input ID.
        :param str label:
            Label.
        :param bool required: (default=False)
            Indicate if this field is required.
        :param str hint: (optional)
            Help message.
        :param str title: (optional)
            Title for the hover effect.
        :param bool clear: (default=False)
            If ``True``, add a ``<div class="cioClear"/>`` at the end.
        :param str class_: (default='cioFormItem')
            The class attribute.
        :param dict attrs:
            Keyworded arguments for ``webhelpers2.html.tags`` object.
        :rtype: str
        :return:
            Output a grid layout.
        """
        # pylint: disable = too-many-arguments
        return grid_item(
            name, label, self.textarea(name, **attrs), required, hint,
            self.error(name), title=title, clear=clear, class_=class_)

    # -------------------------------------------------------------------------
    def _value(self, name, default=None):
        """Return the best value for the field ``name``.

        :param str name:
            Input ID.
        :param str default: (optional)
            Default value.
        :rtype: str
        """
        special = name in self._special[0] or [
            True for k in self._special[1] if name.startswith(k)]
        if not special and name in self._request.POST:
            return self._request.POST[name]
        if name in self.values:
            return self.values[name]
        return default



#!/usr/bin/env python
"""Console script to backup and update Chrysalio sites."""

from sys import executable
from logging import getLogger
from os import chown, makedirs, setgid, setuid, environ
from os.path import join, exists, dirname, expanduser, abspath, normpath
from os.path import relpath
from argparse import REMAINDER, ArgumentParser
from pwd import getpwnam
from getpass import getuser
from compileall import compile_dir
from configparser import ConfigParser
from zipfile import ZIP_DEFLATED, ZipFile

from git import Repo
from git.exc import GitCommandError

from ..lib.i18n import _, translate
from ..lib.utils import walk, scandir
from ..lib.config import config_get, config_get_namespace
from ..lib.log import setup_logging
from ..lib.utils import tounicode, encrypt, decrypt, execute, full_url


LOG = getLogger(__name__)
SECTION_PREFIX = 'Update'


# =============================================================================
def main(args=None):
    """Main function."""
    # Parse arguments
    parser = ArgumentParser('Backup and update Chrysalio sites.')
    parser.add_argument(
        'conf_file', help='configuration file (e.g. cioupdate.ini)')
    parser.add_argument(
        'sites', nargs='*', help='only update these sites (optional)')
    parser.add_argument('--lang', dest='lang', help='user language')
    parser.add_argument(
        '--encrypt', dest='password', help='encrypt password and exit')
    parser.add_argument(
        '--encrypt-key', dest='key', help='optional key for encryption')
    parser.add_argument(
        '--no-backup', dest='no_backup', help='do not backup instance',
        action='store_true')
    parser.add_argument(
        '--no-update', dest='no_update', help='do not update instance',
        action='store_true')
    parser.add_argument(
        '--drop-tables', dest='drop_tables', help='drop existing tables',
        action='store_true')
    parser.add_argument(
        '--remove-locks', dest='remove_locks', action='store_true',
        help='remove locks directory')
    parser.add_argument(
        '--remove-builds', dest='remove_builds', action='store_true',
        help='remove builds directory')
    parser.add_argument(
        '--skip-refresh', dest='skip_refresh', action='store_true',
        help='skip refreshment step')
    parser.add_argument(
        '--recreate-thumbnails', dest='recreate_thumbnails',
        action='store_true', help='recreate thumbnails')
    parser.add_argument(
        '--reindex', dest='reindex', action='store_true',
        help='recreate indexes')
    parser.add_argument(
        '--log-level', dest='log_level', help='log level', default='INFO',
        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'))
    parser.add_argument('--log-file', dest='log_file', help='log file')
    parser.add_argument('extra', nargs=REMAINDER, help='extra options')
    args = parser.parse_args(args)
    conf_file = expanduser(args.conf_file)
    if not exists(conf_file):
        parser.print_usage()
        return
    setup_logging(args.log_level, args.log_file)

    # Backup and update
    CioUpdate(args, conf_file).start(args.sites)


# =============================================================================
class CioUpdate(object):
    """Class to backup and update Chrysalio sites."""

    # -------------------------------------------------------------------------
    def __init__(self, args, conf_file):
        """Constructor method."""
        self._args = args
        self._config = ConfigParser({'here': dirname(abspath(conf_file))})
        self._config.read(tounicode(conf_file), encoding='utf8')
        self._key = args.key or self._config_get('Code', 'key', '-')
        self._cookiecutters = self._config_get('Code', 'cookiecutters')
        if self._cookiecutters:
            self._cookiecutters = expanduser(self._cookiecutters)

    # -------------------------------------------------------------------------
    def start(self, sites=None):
        """Start backup and update.

        :param list sites: (optional)
            Chrysalio sites to backup and update. If empty, this method
            processes all Chrysalio sites defined in configuration file.
        :rtype: bool
        """
        # Only encrypt password
        if self._args.password:
            LOG.info(
                '%s = %s', self._args.password,
                encrypt(self._args.password.encode('utf8'), self._key))
            return True

        # List of sites to update
        if not sites:
            sites = [
                k[len(SECTION_PREFIX):] for k in self._config.sections()
                if k.startswith(SECTION_PREFIX)]
        if 'Code' not in sites:
            sites.insert(0, 'Code')

        # Backup web sites
        is_ok = True
        if not self._args.no_backup:
            LOG.info('{0:=^60}'.format(self._translate(_(' Backup '))))
            for site in sites:
                is_ok &= self._backup(site)
        if not is_ok or self._args.no_update:
            return is_ok

        # Update and compile sources
        LOG.info('{0:=^60}'.format(self._translate(_(' Update '))))
        for site in sites:
            is_ok &= self._update_sources(site)
        if not is_ok:
            return is_ok

        # Populate web sites
        is_ok = True
        LOG.info('{0:=^60}'.format(self._translate(_(' Populating '))))
        for site in sites:
            is_ok &= self._populate(site)
        if not is_ok:
            return is_ok

        # Restart Apache
        if self._config_get('Code', 'reload'):  # pragma: nocover
            LOG.info('{0:-^60}'.format(
                self._translate(_(' Restarting server '))))
            error = execute(self._config_get('Code', 'reload').split())
            LOG.info(self._translate(error[0]))
            if error[1]:
                LOG.error(self._translate(error[1]))
                return False

        return True

    # -------------------------------------------------------------------------
    def _backup(self, site):
        """Backup a Chrysalio site.

        :param str site:
            Name of Chrysalio site to backup.
        :rtype: bool
        """
        # Something to do?
        conf_uri = self._config_get(site, 'conf_uri')
        if conf_uri is None or not exists(conf_uri.partition('#')[0]):
            return True
        directory = self._config_get(site, 'backup.directory')
        command = self._config_get(site, 'backup.command')
        if directory is None or command is None:
            return True
        args = self._config_get(site, 'backup.args')
        user, env = self._user_ids_and_env(self._config_get(site, 'user'))

        # Backup
        LOG.info('%s %s', command, site)
        cmd = [executable, join(dirname(executable), command)]
        cmd = cmd + ['--lang', self._args.lang] if self._args.lang else cmd
        cmd = cmd + ['--log-file', self._args.log_file] \
            if self._args.log_file else cmd
        cmd = cmd + args.split() if args else cmd
        cmd += ['--log-level', self._args.log_level, conf_uri, directory]
        cmd += self._args.extra
        output, error = execute(cmd, preexec_fn=self._demote(*user), env=env)
        if error:
            LOG.error(translate(
                error[33:] or _('"${c}" failed', {'c': command})))
            return False
        if output:
            LOG.warning(output[33:])

        return True

    # -------------------------------------------------------------------------
    def _update_sources(self, site):
        """Update and possibly compile sources.

        :param str site:
            Name of Chrysalio site to backup.
        :rtype: bool
        """
        sources = config_get_namespace(
            self._config, '{0}{1}'.format(SECTION_PREFIX, site), 'source')

        is_ok = True
        for source in sources:
            if not source.endswith('repository'):
                continue
            repository = normpath(self._config_get(
                site, 'source.{0}'.format(source.replace('_', '.'))))
            LOG.info(repository)
            error = self._recursive_chown(repository, getuser())
            if error:  # pragma: nocover
                LOG.error(error)
                is_ok = False
                continue

            error = self._git_pull(repository, site)
            if error:
                LOG.error(self._translate(error))
                is_ok = False
                continue

            code = self._config_get(
                site, 'source.{0}.code'.format(source[:-11]))
            if code is not None and exists(code):
                compile_dir(code, quiet=True)

            cookiecutters = self._config_get(
                site, 'source.{0}.cookiecutters'.format(source[:-11]))
            if self._cookiecutters and cookiecutters is not None \
               and exists(cookiecutters):
                self._zip_cookiecutters(cookiecutters)

            error = self._recursive_chown(
                repository, self._config_get(site, 'user'))
            if error:
                LOG.error(error)
                is_ok = False

        if self._cookiecutters and exists(self._cookiecutters):
            self._recursive_chown(
                self._cookiecutters, self._config_get('Code', 'user'))

        return is_ok

    # -------------------------------------------------------------------------
    def _populate(self, site):
        """Populate a Chrysalio site.

        :param str site:
            Name of Chrysalio site to populate.
        :rtype: bool
        """
        # Something to do?
        conf_uri = self._config_get(site, 'conf_uri')
        if conf_uri is None or not exists(conf_uri.partition('#')[0]):
            return True
        command = self._config_get(site, 'populate.command')
        if command is None:
            return True
        args = self._config_get(site, 'populate.args')
        drop_tables = self._args.drop_tables or \
            self._config_get(site, 'populate.drop_tables') == 'true'
        user, env = self._user_ids_and_env(self._config_get(site, 'user'))
        env['PWD'] = dirname(conf_uri.partition('#')[0])

        # Populate
        LOG.info('###### %s %s ######', command, site)
        cmd = [executable, join(dirname(executable), command)]
        cmd = cmd + ['--drop-tables'] if drop_tables else cmd
        cmd = cmd + ['--remove-locks'] if self._args.remove_locks else cmd
        cmd = cmd + ['--remove-builds'] if self._args.remove_builds else cmd
        cmd = cmd + ['--skip-refresh'] if self._args.skip_refresh else cmd
        cmd = cmd + ['--recreate-thumbnails'] \
            if self._args.recreate_thumbnails else cmd
        cmd = cmd + ['--reindex'] if self._args.reindex else cmd
        cmd = cmd + ['--lang', self._args.lang] if self._args.lang else cmd
        cmd = cmd + ['--log-file', self._args.log_file] \
            if self._args.log_file else cmd
        cmd = cmd + args.split() if args else cmd
        cmd += ['--log-level', self._args.log_level, conf_uri]
        cmd += self._args.extra
        output, error = execute(
            cmd, cwd=env['PWD'], preexec_fn=self._demote(*user), env=env)
        if error:
            LOG.error(translate(output or _('"${c}" failed', {'c': command})))
            return False
        LOG.info(output[33:])

        return True

    # -------------------------------------------------------------------------
    def _git_pull(self, repo_path, site):
        """Pull and update a Git repository.

        :param str repo_path:
            Absolute path to of the repository.
        :param str site:
            Name of Chrysalio site to backup.
        :rtype: pyramid.i18n.TranslationString
        :return:
            Error message or ``None``.
        """
        if not exists(join(repo_path, '.git', 'config')):
            return _('This directory is not a Git repository.')

        gituser = self._config_get(site, 'gituser', 'gituser')
        gitpassword = decrypt(
            self._config_get(site, 'gitpassword', ''), self._key)
        repo = Repo(repo_path)
        if not repo.remotes:  # pragma: nocover
            return _('No remote repository')
        url = repo.remotes.origin.url
        fullurl = full_url(url, gituser, gitpassword)
        try:
            repo.remotes.origin.set_url(fullurl)
            repo.remotes.origin.pull()
        except GitCommandError as error:  # pragma: nocover
            return str(error).replace(fullurl, url)
        finally:
            repo.remotes.origin.set_url(url)
        return None

    # -------------------------------------------------------------------------
    @classmethod
    def _recursive_chown(cls, path, user):
        """Change owner of directory ``path`` recursively.

        :param str path:
            Absolute path to directory to process.
        :param str user:
            User name.
        :rtype: str
        :return:
            Error message or ``None``.
        """
        if not user:
            return None
        try:
            user = getpwnam(user)
        except KeyError as error:
            return error

        try:
            chown(path, user.pw_uid, user.pw_gid)
        except OSError as error:  # pragma: nocover
            return error

        for root, dirs, files in walk(path.encode('utf8')):
            for name in dirs:
                if name == b'.tox':
                    dirs.remove(name)  # pragma: nocover
                else:
                    chown(join(root, name), user.pw_uid, user.pw_gid)
            for name in files:
                chown(join(root, name), user.pw_uid, user.pw_gid)
        return None

    # -------------------------------------------------------------------------
    def _zip_cookiecutters(self, cookiecutters):
        """Zip each cookiecutter found in ``cookiecutters`` directory.

        :param str cookiecutters:
            Absolute path to cookiecutters directory.
        """
        if not exists(self._cookiecutters):
            makedirs(self._cookiecutters)

        for entry in scandir(cookiecutters):
            if not exists(join(entry.path, 'cookiecutter.json')):
                continue  # pragma: nocover
            filename = join(self._cookiecutters, '{0}.zip'.format(entry.name))
            with ZipFile(filename, 'w', ZIP_DEFLATED) as zip_file:
                root = normpath(join(entry.path, '..'))
                zip_file.write(entry.path, entry.name)
                for path, unused_, files in walk(entry.path):
                    for name in files:
                        name = join(path, name)
                        zip_file.write(name, relpath(name, root))
            LOG.info(self._translate(
                _('${n} -> ${f}', {'n': entry.name, 'f': filename})))

    # -------------------------------------------------------------------------
    @classmethod
    def _user_ids_and_env(cls, user=None):
        """Return a tuple such as ``((uid, group), env)``.

        :param str user: (optional)
            User log name.
        :rtype: tuple
        """
        log_name = user or getuser()
        try:
            pw_record = getpwnam(log_name)
        except KeyError:
            pw_record = getpwnam(getuser())

        env = environ.copy()
        env.update({
            'HOME': pw_record.pw_dir, 'LOGNAME': pw_record.pw_name,
            'USER': pw_record.pw_name})
        return (pw_record.pw_uid, pw_record.pw_gid), env

    # -------------------------------------------------------------------------
    @classmethod
    def _demote(cls, user_uid, user_gid):
        """Demote the current user.

        :param int user_uid:
            User UID.
        :param int user_gid:
            User group.
        """
        def result():  # pragma: nocover
            """Change user."""
            setgid(user_gid)
            setuid(user_uid)
        return result

    # -------------------------------------------------------------------------
    def _config_get(self, section, option, default=None):
        """Retrieve a value from the `cioupdate` configuration file.

        :param str section:
            Section name.
        :param str option:
            Option name.
        :param str default: (optional)
            Default value
        :rtype: str
        """
        return config_get(
            self._config, '{0}{1}'.format(SECTION_PREFIX, section), option,
            default)

    # -------------------------------------------------------------------------
    def _translate(self, text):
        """Return ``text`` translated.

        :param str text:
            Text to translate.
        :rtype: str
        """
        return translate(text, self._args.lang)



# pylint: disable = import-outside-toplevel
"""Tests of ``views.backup`` class."""

from os import listdir, makedirs
from os.path import exists, join
from collections import OrderedDict, namedtuple
from cgi import FieldStorage

from pyramid.response import Response, FileResponse

from . import DBUnitTestCase


# =============================================================================
class UViewsBackupBackupView(DBUnitTestCase):
    """Unit test class for testing :class:`views.backup.BackupView`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""

        super(UViewsBackupBackupView, self).setUp()

        self.request.matched_route = namedtuple('Route', 'name')(name='backup')
        self.configurator.registry['modules'] = OrderedDict()
        self.configurator.registry['modules_off'] = set()
        self.configurator.registry.settings['site.uid'] = 'test'

    # -------------------------------------------------------------------------
    def test_index(self):
        """[u:views.backup.BackupView.index]"""
        from ..views.backup import BackupView

        # Page display
        response = BackupView(self.request).index()
        self.assertIn('form', response)
        self.assertIn('action', response)
        self.assertIn('download_max_size', response)
        self.assertEqual(
            self.request.breadcrumbs.current_title(), 'Configuration Backup')

    # -------------------------------------------------------------------------
    def test_index_backup(self):
        """[u:views.backup.BackupView.index] Backup"""
        from ..views.backup import BackupView
        from . import ATTACHMENTS_DIR

        # Backup without attachments
        self.request.POST = {'bck!.x': True}
        response = BackupView(self.request).index()
        self.assertIsInstance(response, Response)

        # Backup with attachments
        self.configurator.registry.settings['attachments'] = ATTACHMENTS_DIR
        response = BackupView(self.request).index()
        self.assertIsInstance(response, FileResponse)

    # -------------------------------------------------------------------------
    def test_index_restore(self):
        """[u:views.backup.BackupView.index] Restore"""
        from . import TEST_INI, TEST1_INI, TEST1_USR_ZIP, TEST2_USR_ZIP
        from . import BACKUP_DIR, DummyPOST
        from ..models.dbuser import DBUser

        dbsession = self.request.dbsession
        self.configurator.add_route('logout', '/logout')
        self.request.POST = DummyPOST()
        self.request.POST['rst!.x'] = True

        # Without a main administrator
        self.request.params['reset'] = True
        self.request.registry.settings['__file__'] = TEST1_INI
        response = self._call_index_with_file(TEST1_USR_ZIP)
        self.assertIsInstance(response, dict)
        self.assertTrue(self.request.session.pop_flash('alert'))

        # With an incorrect backup
        self.request.registry.settings['__file__'] = TEST_INI
        response = self._call_index_with_file(TEST2_USR_ZIP)
        self.assertIsInstance(response, dict)
        self.assertTrue(self.request.session.pop_flash('alert'))

        # A correct backup with reset
        makedirs(BACKUP_DIR)
        self.request.registry.settings['attachments'] = BACKUP_DIR
        response = self._call_index_with_file(TEST1_USR_ZIP)
        self.assertNotIn('user', self.request.session)
        self.assertFalse(self.request.session.pop_flash('alert'))
        self.assertFalse(self.request.session.pop_flash())
        self.assertEqual(len(dbsession.query(DBUser).all()), 2)
        self.assertIsNotNone(
            dbsession.query(DBUser).filter_by(login='admin').first())
        self.assertIsNotNone(
            dbsession.query(DBUser).filter_by(login='user3').first())
        self.assertTrue(listdir(BACKUP_DIR))
        self.assertTrue(exists(join(
            BACKUP_DIR, DBUser.attachments_dir, 'User3', 'user3.svg')))

        # A correct backup without reset
        self.request.params['reset'] = False
        response = self._call_index_with_file(TEST1_USR_ZIP)
        self.assertFalse(self.request.session.pop_flash('alert'))
        self.assertTrue(self.request.session.pop_flash())

        # Ajax
        self.request.is_xhr = True
        response = self._call_index_with_file(TEST1_USR_ZIP)
        self.assertIsInstance(response, dict)
        self.assertTrue(self.request.session.pop_flash())

    # -------------------------------------------------------------------------
    def _call_index_with_file(self, filename):
        """Call index with a file.

        :param str filename:
            Absolute path to the file to use.
        :rtype: dict
        """
        from ..views.backup import BackupView

        with open(filename, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = filename
            input_file.file = hdl
            self.request.POST.multikeys = {'file': (input_file,)}
            response = BackupView(self.request).index()
        return response



# pylint: disable = import-outside-toplevel
"""Tests of ``models`` functions and class."""

from unittest import TestCase

from sqlalchemy import Column, Integer
from sqlalchemy.orm.session import sessionmaker, Session
from sqlalchemy.engine.base import Engine
import transaction

from pyramid import testing

from . import DBUnitTestCase


# =============================================================================
class UModelsIncludeme(TestCase):
    """Unit test class for :func:`models.includeme`."""

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects of ``pyramid.testing.setUp()``."""
        from ..models import DB_METADATA

        testing.tearDown()
        transaction.manager.abort()
        if DB_METADATA.bind:
            DB_METADATA.drop_all()

    # -------------------------------------------------------------------------
    def test_not_defined(self):
        """[u:models.includeme] not defined"""
        from ..models import includeme

        configurator = testing.setUp()
        self.assertRaises(SystemExit, includeme, configurator)

    # -------------------------------------------------------------------------
    def test_defined(self):
        """[u:models.includeme] defined"""
        from ..models import DB_METADATA

        configurator = testing.setUp(settings={
            'sqlalchemy.url': 'sqlite:///:memory:'})
        configurator.include('..models')
        self.assertIn('dbsession_factory', configurator.registry)
        self.assertIsNotNone(configurator.registry['dbsession_factory'])
        self.assertIsInstance(
            configurator.registry['dbsession_factory'], sessionmaker)
        self.assertIsNotNone(DB_METADATA.bind)
        self.assertIsInstance(DB_METADATA.bind, Engine)


# =============================================================================
class UModelsGetDBEngine(TestCase):
    """Unit test class for :func:`models.get_dbengine`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.get_dbengine]"""
        from ..models import get_dbengine

        dbengine = get_dbengine({'sqlalchemy.url': 'sqlite:///:memory:'})
        self.assertIsInstance(dbengine, Engine)


# =============================================================================
class UModelsGetDBSessionFactory(TestCase):
    """Unit test class for :func:`models.get_dbsession_factory`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.get_dbsession_factory]"""
        from ..models import get_dbengine, get_dbsession_factory

        dbsession_factory = get_dbsession_factory(
            get_dbengine({'sqlalchemy.url': 'sqlite:///:memory:'}))
        self.assertIsInstance(dbsession_factory, sessionmaker)


# =============================================================================
class UModelsGetTMDBSession(DBUnitTestCase):
    """Unit test class for :func:`models.get_tm_dbsession`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.get_tm_dbsession]"""
        from ..models import get_tm_dbsession

        with transaction.manager:
            dbsession = get_tm_dbsession(
                self.configurator.registry['dbsession_factory'],
                transaction.manager)
        self.assertIsInstance(dbsession, Session)


# =============================================================================
class UModelsAddColumn(DBUnitTestCase):
    """Unit test class for :func:`models.add_column`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:models.add_column]"""
        from ..models import add_column
        from ..models.dbuser import DBUser

        # pylint: disable = no-member
        counter = len(DBUser.__table__.columns)
        add_column(DBUser, Column('Age', Integer))
        self.assertEqual(len(DBUser.__table__.columns), counter + 1)



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.clipboard`` class."""

from unittest import TestCase

from pyramid.testing import DummyRequest

from . import ConfiguratorTestCase


# =============================================================================
class UIncludesClipboardIncludeme(ConfiguratorTestCase):
    """Unit test class for :func:`includes.clipboard.includeme`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.clipboard.includeme]"""
        self.configurator.include('..includes.clipboard')
        self.assertIn('panels', self.configurator.registry)
        self.assertIn('clipboard', self.configurator.registry['panels'])


# =============================================================================
class ULibClipboardlClipboard(TestCase):
    """Unit test class for :class:`lib.clipboard.Clipboard`."""

    # -------------------------------------------------------------------------
    def test_is_empty(self):
        """[u:lib.clipboard.Clipboard.is_empty]"""
        from ..includes.clipboard import Clipboard

        # Empty
        request = DummyRequest()
        self.assertTrue(Clipboard.is_empty(request))

        # Not empty
        request.session['clipboard'] = [('wfile', False, 'my_data', 'My data')]
        self.assertFalse(Clipboard.is_empty(request))

    # -------------------------------------------------------------------------
    def test_entries(self):
        """[u:lib.clipboard.Clipboard.entries]"""
        from ..includes.clipboard import Clipboard

        request = DummyRequest()
        request.session['clipboard'] = [
            ('wfile', False, 'my_data1', 'My data 1'),
            ('wfile', False, 'my_data2', 'My data 2')]
        entries = Clipboard.entries(request)
        self.assertEqual(len(entries), 2)

    # -------------------------------------------------------------------------
    def test_push(self):
        """[u:lib.clipboard.Clipboard.push]"""
        from ..includes.clipboard import Clipboard

        request = DummyRequest()
        request.registry = {'settings': {'clipboard-size': 3}}
        Clipboard.push(request, 'wfile', False, 'my_data', 'My data')
        entries = Clipboard.entries(request)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0][0], 'wfile')
        self.assertFalse(entries[0][1])
        self.assertEqual(entries[0][2], 'my_data')
        self.assertEqual(entries[0][3], 'My data')

    # -------------------------------------------------------------------------
    def test_selection(self):
        """[u:lib.clipboard.Clipboard.selection]"""
        from ..lib.panel import PANEL_ITEM_PREFIX
        from ..includes.clipboard import Clipboard

        # Empty clipboard
        request = DummyRequest()
        request.registry = {'settings': {'clipboard-size': 3}}
        self.assertFalse(Clipboard.selection(request, ('wfile',)))

        # Without explicit request
        Clipboard.push(request, 'wfile', True, 'my_data1', 'My data 1')
        Clipboard.push(request, 'wfile', False, 'my_data2', 'My data 2')
        Clipboard.push(request, 'user', False, 4, 'Luc SAMBOUR')
        selection = Clipboard.selection(request, ('wfile',))
        self.assertEqual(len(selection), 1)
        self.assertEqual(selection[0][2], 'my_data2')
        entries = Clipboard.entries(request)
        self.assertEqual(len(entries), 3)
        self.assertEqual(entries[0][2], 'my_data2')
        self.assertEqual(entries[1][2], 4)
        self.assertEqual(entries[2][2], 'my_data1')

        # With explicit request
        request.params = {
            '{0}1'.format(PANEL_ITEM_PREFIX): True,
            '{0}2'.format(PANEL_ITEM_PREFIX): True}
        selection = Clipboard.selection(request, ('wfile',))
        self.assertEqual(len(selection), 1)
        self.assertEqual(selection[0][2], 'my_data1')
        entries = Clipboard.entries(request)
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0][2], 'my_data2')
        self.assertEqual(entries[1][2], 4)

        # Too many enntries
        Clipboard.push(request, 'wfile', True, 'my_data3', 'My data 3')
        Clipboard.push(request, 'wfile', True, 'my_data4', 'My data 4')
        entries = Clipboard.entries(request)
        self.assertEqual(len(entries), 3)
        self.assertEqual(entries[0][2], 'my_data4')
        self.assertEqual(entries[1][2], 'my_data3')
        self.assertEqual(entries[2][2], 'my_data2')



"""Base class for view callables."""

from ..lib.panel import Panel


# =============================================================================
class BaseView(object):
    """Base class to manage with panel.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    # pylint: disable = too-few-public-methods

    # -------------------------------------------------------------------------
    def __init__(self, request):
        """Constructor method."""
        self._request = request
        Panel.manage_panels(request)
